//
//  TechFactory.hu
//  Szenzor + kijelzés minta 
//  Egy gombbal választható melyik szenzor értékét mutassa az LCD+I2C kijelző
//  Oktatós #5
//

#include <Wire.h> 

//LiquidCrystal_I2C lib, telepíthető a könyvtárkezelőből is
//https://github.com/fdebrabander/Arduino-LiquidCrystal-I2C-library
//version: 1.1.2
#include <LiquidCrystal_I2C.h>

//Adafruit BME280 lib, telepíthető a könyvtárkezelőből is
//https://github.com/adafruit/Adafruit_BME280_Library
//version: 1.0.8
//szükséges még az Adafruit Unified Sensor (v1.0.3), telepíthető a könyvtárkezelőből vagy
//https://github.com/adafruit/Adafruit_Sensor
#include <Adafruit_BME280.h>

//Magyar ékezetes karakterek támogatása az LCD kijelzőn
//https://github.com/TechFactoryHU/LCDCharConverter
#include "LCDCharConverter.h"

//elso parameter az eszkoz cime
//ha nem talalod az eszkoz cimet, hasznald az i2c scanner kodot
//https://playground.arduino.cc/Main/I2cScanner/
LiquidCrystal_I2C lcd(0x27,16,2);
LCDCharConverter converter(&lcd);

//BME szenzor init, i2c kommunikációval
Adafruit_BME280 bme; // I2C

//A tengerszint feletti magasság méréséhez szükség van a tengerszinti légnyomásra
//ezt az értéket le kellene minden nap kérdezni a lehető legpontosabb  tengerszint feletti magasság számításhoz
//így a konstans beállításával ~10m -es pontosságra kell számítani.
#define SL_PRESSURE 1013.25

//IR jelenlét érzékelő pinje
#define IR_PIN D6
//Gomb pinje
#define BTN_PIN D5

//szenzorok száma
#define SENSORS_COUNT 5
//szenzor struct; tartalmazza a nevet, pld Hőmérséklet, 
//a mértékegységet és az értéket. Ennek segítségével iratjuk ki a kijelzőre a mért adatokat
//van még egy típusa is, ami meghatározza, hogy melyik szenzort értékét kell a value változóba beolvasni
//a beolvasást a readSensors() funkció végzi
struct Sensor {
    char title[16];   //kiirandó név, pld Hőmérséklet, lcd 1. sorába íródik. 
    char postfix[5];  //mértékegység, pld °C, vagy hPa
    uint8_t type;     //Szenzor típusa; ezzel tudatjuk, hogy melyik szenzort kell a value változóba beolvasni 
    uint8_t decimals; //value érték formázása (0=egész szám, 1=tized, 2=század ...)
    uint8_t display;  //megjelenés típusa (0=szám adat, 1=be/ki szenzor állapot)
    float value;      //Utoljára kiolvasott érték
    bool changed;     //ezzel a bool változóval tudatjuk, hogyha egy érték változott a korábbihoz képest
};

//Szenzor tömb felépítése
Sensor sensors[SENSORS_COUNT] = {
  {"Hőmérséklet", "°C", 0, 2, 0},
  {"Páratartalom", "%", 1, 1, 0},
  {"Légnyomás", "hPa", 2, 2, 0},
  {"Tszf. magasság", "m", 3, 1, 0},
  {"IR szenzor", "", 4, 0, 1}
};

//jelenleg kiiratott aktív szenzor;
uint8_t activeSensor = 0;
//a kijelzőre kiírt adatok változásakor true-ra váltjuk, ha megtörtént a kijelző frissítése ismét false lesz
bool displayChanged = false;
bool backlightStatus = false;

//gomb utolsó állapotát figyeljük vele
uint8_t _btnlaststatus = 0;
unsigned long _lastbtnchange;
//a gomb utolsó lenyomásától számított 15 mp után indulhat el az automatikus szenzor érték figyelés/váltás
const int _buttonstandbyms = 15000; 

//a kijelző utolsó változását figyeljük vele, ehhez mérten kapcsoljuk le majd a háttérvilágítást
unsigned long _lastchange;
//háttérvilágítás lekapcsolásának ideje, ha nem történik változás
const int _backlightoffinterval = 10000; //10 mp múlva kapcsoljon le
const int _autochangeinterval = 5000; // 5mp

/*
 *  --- SETUP ---
 */
 
void setup() {
  Serial.begin(115200);
  Serial.println("Starting up");
  converter.reset();
  
  //aktiváljuk a belső felhúzó ellenállást, így a gombra már nem kell külön ellenállás
  //ilyenkor a GND-re kell kötni a gomb másik oldalát
  //a tech sheetben tudod megnézni, hogy melyik pinek támogatottak
  pinMode(BTN_PIN, INPUT_PULLUP);
  pinMode(IR_PIN, INPUT);
  
  //LCD beállítása 16x2-re (16 karakter széles és 2 soros)
  lcd.begin(16,2);
  lcd.init();     
  //háttérvilágítás bekapcsolása
  setBacklight(true);
  
  if (!bme.begin(0x76)) {
      Serial.println("Nem található a BME280 szenzor, ellenőrízd a kábeleket!");
      lcd.print("BME280 nincs");
      lcd.setCursor(0,1);
      lcd.print("csatlakoztatva!");
      while (1);
  }
  
  converter.display("MérőEgységMinta");
  lcd.setCursor(0,1);
  lcd.print("-TechFactory.hu-");
  delay(5000);
  lcd.clear();
  displayChanged = true;
  activeSensor = 0;
}


/*
 *  --- LOOP ---
 */
 
void loop() {
    //beolvassuk az összes szenzor értékét
    readSensors();

    //gomb interakcióval válhatjuk, hogy mit jelezzen ki
    checkButton();

    //ha valamelyik szenzor változott, automatikusan oda léptetjük a kijelzőt (és bekapcsoljuk a háttérvilágítást)
    //csak abban az esetben futtatjuk az "automatikus" váltás ha az eszköz gombját nem használjuk épp
    //ehhez szintén egy unsigned long változót használunk + egy interval amit a gomb utolsó lenyomásától számolunk
    //plusz, figyeljük, hogy legalább 1x már változott e a kijelző, így első indításnál nem fut bele ide egyszerre
    if (_lastchange > 0 && _lastbtnchange + _buttonstandbyms < millis()) {
        if (_lastchange + _autochangeinterval < millis()) {
            int chg = getLastChangedSensor();
            if (chg > -1) {
                activeSensor = chg;
                displayChanged = true;
            }
        }
    }
    
    if (displayChanged) {
        //háttérvilágítás be
        setBacklight(true);
        lcd.clear(); //törlünk minden korábban kiíratott adatot
        //a kijelző első sorába kiíratjuk a mért adat nevét
        converter.display(sensors[activeSensor].title, 0, 0);

        String value;
        if (sensors[activeSensor].display==0) {
            if (sensors[activeSensor].decimals > 0) {
                value += String(roundTo(sensors[activeSensor].value, pow(10, sensors[activeSensor].decimals)));
            }else {
                value += String((int)round(sensors[activeSensor].value));
            }
            value += " "+String(sensors[activeSensor].postfix);
        }else if(sensors[activeSensor].display==1) {
            value = sensors[activeSensor].value == 1 ? "BE = 1" : "KI = 0";
        }

        //nullázzuk a változást
        sensors[activeSensor].changed = false;
        converter.display(&value[0], 0, 1);
        //frissítjük a kijelző utolsó változásának idejét
        _lastchange = millis();
        displayChanged = false;
    }

    //LCD háttérvilágításának auto kikapcsolása, ha eltelt a megadott idő
    if (backlightStatus && _lastchange + _backlightoffinterval < millis()) {
        setBacklight(false);
    }
}


/*
 *  --- EGYÉB FUNKCIÓK ---
 */

//helper funkcio, mivel az LCD libtől nem lehet lekérdezni a 
//háttérvilágítás állapotát, ezért nekünk kell róla gondoskodni
void setBacklight(bool s) {
    lcd.setBacklight(s?1:0);
    backlightStatus = s;
}

//a gomb állapotváltozását figyeljük ezzel a funkcióval
//használható az attachInterrupt is, viszont nem minden pinen támogatott, erre érdemes figyelni
void checkButton() {
    if (digitalRead(BTN_PIN) != _btnlaststatus) {
       _btnlaststatus = digitalRead(BTN_PIN);
       //gomb lenyomva (mivel INPUT_PULLUP be van kapcsolva ezért ha nincs lenyomva 1-et ad vissza)
       if (_btnlaststatus == 0) {
          //ha nem aktív a háttérvilágítás, akkor bekapcsoljuk először
          if (!backlightStatus) {
               _lastchange = millis();
              setBacklight(true);
              return;
          }else {
              //növeljük az activeSensor változót, ha körbeért akkor 0-ra váltjuk
              activeSensor++;
              if (activeSensor >= SENSORS_COUNT) { activeSensor = 0; }
              //tudatjuk a kijelzővel, hogy frissíteni kell a kiírtakat
              displayChanged = true;
          }
       //gomb felengedve
       }else { }
       //gomb utolsó állapotváltozásának rögzítése
       _lastbtnchange = millis();
    }
}


//egy adott szenzor értékének (újra) beolvasása
void readSensor(uint8_t i) {
  float sens = 0;
  if (i>=0 && i<SENSORS_COUNT) {
      //az eltárolt típus alapján döntjük el, hogy melyik értéket melyik szenzorról kell leolvasni
      switch (sensors[i].type) {
          case 0:
              sens = bme.readTemperature(); 
              break;
          case 1:
              sens = bme.readHumidity(); 
              break;
          case 2:
              //légköri nyomás Pa-ban, nekünh hPa kell
              sens = bme.readPressure() / 100; 
              break;
          case 3:
              sens = bme.readAltitude(SL_PRESSURE); 
              break;
          case 4:
              //az ir szenzor fordított logikával dolgozik; ha nincs mozgás akkor 1 ha van akkor pedig 0-ad, ezt itt megfordítjuk.
              sens = digitalRead(IR_PIN) ? 0 : 1;
              break;
      }

      bool chgd = false;
      if (sensors[i].decimals > 0) {
          sens = roundTo(sens,  pow(10, sensors[i].decimals));
          if (sens - sensors[i].value > 0.5) {
             chgd = true;
          }
      }else {
          sens = round(sens);
          if (sensors[i].value != sens)  {
              chgd = true;
          }
      }

      if (chgd) {
         Serial.print(sensors[i].title);
         Serial.print(" = ");
         Serial.println(sens);
         sensors[i].value = sens;
         sensors[i].changed = true;
         if (i == activeSensor) {
             displayChanged = true;
         }
      }
  }
}

//az összes szenzor értékének beolvasása
void readSensors() {
   for (int i=0; i<SENSORS_COUNT; i++) {
       readSensor(i);
   }
}

//visszaadja, hogy melyik szenzor értéke változott, -1 ha egyiké sem
int getLastChangedSensor() {
    for (int i=0; i<SENSORS_COUNT; i++) {
       if (sensors[i].changed) { return i; }
    }
    return -1;
}

//nullázzuk a "changed" értékeket
void resetSensors() {
   for (int i=0; i<SENSORS_COUNT; i++) {
      sensors[i].changed = false;
   }
}

//lebegőpontos számok kerekítésére a megadott tizedes helyre 
float roundTo(float n, int d) {
   return (float)((int)(n*d))/d;
}
